* WordTables.prg		(c) 2002 Peisch Custom Software, Inc.
* www.peisch.com

*   Reads data from tables in a Word document and loads the contents into an array property

*    Freeware utility - may be distributed royalty-free.

* 07/08/02 - Added WriteTable method
* 05/11/02 - Original version

* Properties:
*   oDocument = Object reference to Word document, if you already have it open.
*               OK to leave this property as .null. if you haven't opened a document.
*   cWordFile = Name of Word document to open.  If not filled in at the time
*               the ReadTable method is called and oDocument is .null., a GETFILE 
*               dialog will be opened for the user to select a file at that time.
*               If none selected, processing will terminate, and the cErrorMessage
*               property will be populated.
*   nTable = Specifies which table from the document to read (using the Tables collection
*            of the Word document).  Default is 1.
*   nStartRow = Starting row to read from table.  Default is 1.
*   nEndRow = Ending row to read from table.  Default is end last row.
*   nStartCol = Starting column to read from table.  Default is 1.
*   nEndCol = Ending column to read from table.  Default is last column.
*   aTableData = Array that is loaded with data from table.
*   nErrorNum = VFP error number, if an error encountered.
*   cErrorMessage = Either a VFP error message, if a VFP error is encountered, or an
*                   error message from this program, if some other kind of error is encountered.
*   lShowStatus = Set to .T. if you want a WAIT WINDOW message showing the current row and column
*                 being processed.  Default is .F.
*   lRemoveUnprintable = Set to .T. if you want any unprintable characters removed from each cell. Default is .F.
*   lErrorIfSmaller = When writing, set to .T. if you want to return an error if the rows or columns in 
*                     aTableData, plus the nStartRow/nStartCol exceed the existing table's size. Default is .F.
*                     If .F. and the table is too small, the table will be expanded to hold the extra data.
*   lCreateIfMissing = When writing, set to .F. if you don't want a table created in cases where nTable doesn't exist.
*                      Default is .T.  The new table will be created at the end of the document.

* USAGE:
*  To Read a Table:
*!*	  oWordTables = NEWOBJECT("WordTables","WordTables.prg")

*!*	  * Technique 1:
*!*	  * Let the user open the table, and use all the defaults:
*!*	  oWordTables.ReadTable()
*!*	  IF EMPTY(oWordTables.cErrorMessage)
*!*	     * The table was read, and the contents can be accessed
*!*	     * by reading oWordTables.aTableData
*!*	  ELSE 
*!*	     * There was an error.
*!*	  ENDIF 

*!*	  * Technique 2:
*!*	  * Specify the table, but let WordTables open it. Use all the defaults:
*!*	  oWordTables.cWordFile = "C:\FullPath\SomeWordFile.doc"
*!*	  oWordTables.ReadTable()
*!*	  IF EMPTY(oWordTables.cErrorMessage)
*!*	     * The table was read, and the contents can be accessed
*!*	     * by reading oWordTables.aTableData
*!*	  ELSE 
*!*	     * There was an error.
*!*	  ENDIF 

*!*	  * Technique 3:
*!*	  * Start with a document that's already open, and specify defaults:
*!*	  oWordTables.oDocument = oMyOpenDocument
*!*	  oWordTables.nTable = 2
*!*	  oWordTables.nStartRow = 3
*!*	  oWordTables.nEndRow = 10
*!*	  oWordTables.nStartCol = 5
*!*	  oWordTables.nEndCol = 8
*!*	  oWordTables.ReadTable()
*!*	  IF EMPTY(oWordTables.cErrorMessage)
*!*	     * The table was read, and the contents can be accessed
*!*	     * by reading oWordTables.aTableData
*!*	  ELSE 
*!*	     * There was an error.
*!*	  ENDIF 

* To Write a Table:
*   As with reading tables, there are three possible techniques.
*   We'll just use technique 1 here.  The important thing is that you must
*   populate the aTableData array after instantiating the WordTables object.
*!*	  oWordTables = NEWOBJECT("WordTables","WordTables.prg")
*!*	   DIMENSION oWordTables.aTableData[2,2]
*!*	   oWordTables.aTableData[1,1] = "Apples"
*!*	   oWordTables.aTableData[1,2] = "Red"
*!*	   oWordTables.aTableData[2,1] = "Bananas"
*!*	   oWordTables.aTableData[2,2] = "Yellow"
*!*	   oWordTables.WriteTable()
*!*	  IF EMPTY(oWordTables.cErrorMessage)
*!*	     * Success!
*!*	  ELSE 
*!*	     * There was an error.
*!*	  ENDIF 

DEFINE CLASS WordTables as Relation olepublic

DIMENSION aTableData[1]
aTableData = ''
oDocument = .null.
cWordFile = ''
nTable = 1
nStartRow = 1
nEndRow = 0
nStartCol = 1
nEndCol = 0
nErrorNum = 0
nErrorLine = 0
cErrorMessage = ''
nErrorLine = 0
lShowStatus = .F.
lRemoveUnprintable = .F.
lErrorIfSmaller = .F.
lCreateIfMissing = .T.

*----------------------------------------------------------------------------------
FUNCTION ReadTable
LOCAL loWord, loDocument, lnLastRow, lnLastCol, lnI, lnJ, llWasOpen, lnError, lcOldOnError, llError, lnCurRow, lnCurCol

lcOldOnError = ON("Error")
lnError = 0
ON ERROR lnError = ERROR()
llError = .F.

* Clear out the array before we begin, in case there's something left from
* a previous call.
DIMENSION this.aTableData[1]
this.aTableData = ''
this.nErrorNum = 0
this.cErrorMessage = ''

IF ISNULL(this.oDocument)
	IF EMPTY(this.cWordFile)
		this.cWordFile = GETFILE("doc")
		IF EMPTY(this.cWordFile)
			this.cErrorMessage = "No Word document specified"
			llError = .T.
		ENDIF
	ENDIF
	
	IF NOT llError AND NOT FILE(this.cWordFile)
		ASSERT .f. message this.cWordFile+" not found!"
		this.cErrorMessage = this.cWordFile+' not found!'
		llError = .T.
	ENDIF

	IF NOT llError
		loWord = CREATEOBJECT("Word.Application")
		IF VARTYPE(loWord) <> "O"
			ASSERT .F. MESSAGE "Couldn't instantiate Word"
			this.cErrorMessage = "Could not instantiate Word"
			llError = .T.
		ENDIF
		IF NOT llError
			loDocument = loWord.Documents.Open(this.cWordFile)
			IF VARTYPE(loDocument) <> "O"
				ASSERT .F. MESSAGE "Couldn't open document"
				this.cErrorMessage = "Could not open document"
				llError = .T.
			ENDIF 
		ENDIF 
		llWasOpen = .F.
	ENDIF
ELSE
	loDocument = this.oDocument
	llWasOpen = .T.
ENDIF

IF NOT llError AND lnError = 0 AND VARTYPE(loDocument.Tables[this.nTable]) <> 'O'
	ASSERT .F. message "Cannot find table number "+TRANSFORM(this.nTable)+" in document!"
	this.cErrorMessage = "Cannot find table number "+TRANSFORM(this.nTable)+" in document!"
	llError = .T.
ENDIF

IF NOT llError AND lnError = 0
	lnLastRow = loDocument.Tables[this.nTable].Rows.Count
	lnLastCol = loDocument.Tables[this.nTable].Columns.Count

	DO CASE
	CASE this.nEndRow = 0
		this.nEndRow = lnLastRow
	CASE this.nEndRow > lnLastRow
		this.nEndRow = lnLastRow
	OTHERWISE
		* property is OK
	ENDCASE
		
	DO CASE
	CASE this.nEndCol = 0
		this.nEndCol = lnLastCol
	CASE this.nEndCol > lnLastCol
		this.nEndCol = lnLastCol
	OTHERWISE
		* property is OK
	ENDCASE

	lnI = this.nEndRow - this.nStartRow + 1
	lnJ = this.nEndCol - this.nStartCol + 1
	DIMENSION this.aTableData[lnI,lnJ]

	FOR lnI = this.nStartRow TO this.nEndRow
		FOR lnj = this.nStartCol TO this.nEndCol
			IF this.lShowStatus
				WAIT WINDOW NOWAIT "Reading row "+TRANSFORM(lnI)+" of "+TRANSFORM(this.nEndRow)+ CHR(13) ;
						+"Column "+TRANSFORM(lnJ)+" of "+TRANSFORM(this.nEndCol)
			ENDIF 
			
			lnCurRow = lnI - this.nStartRow + 1
			lnCurCol = lnJ - this.nStartCol + 1
			
			WITH loDocument.Tables[this.nTable].Rows[lnI]
				* Remove the end-of-cell marker from the cell
				.Cells[lnJ].Range.End = .Cells[lnJ].Range.End - 1
				IF lnError <> 0
					EXIT 
				ENDIF 
				this.aTableData[lnCurRow,lnCurCol] = ALLTRIM(.Cells[lnJ].Range.Text)
				IF this.lRemoveUnprintable
					this.aTableData[lnCurRow,lnCurCol] = this.RemoveUnprintable(this.aTableData[lnCurRow,lnCurCol])
				ENDIF 
				IF lnError <> 0
					EXIT 
				ENDIF 
			ENDWITH
		ENDFOR
		IF lnError <> 0
			EXIT 
		ENDIF 
	ENDFOR
	IF this.lShowStatus
		WAIT CLEAR
	ENDIF 
ENDIF

IF NOT llWasOpen
	IF VARTYPE(loDocument) = "O"
		loDocument.Close()
	ENDIF
	IF VARTYPE(loWord) = "O"
		loWord.Quit()
	ENDIF 
ENDIF 

ON ERROR &lcOldOnError
this.nErrorNum = lnError
IF lnError <> 0
	this.cErrorMessage = MESSAGE()
ENDIF 

ENDFUNC
*----------------------------------------------------------------------------------
FUNCTION RemoveUnprintable
* Remove unprintable characters from a string
LPARAMETERS tcString
LOCAL lnI, lcReturn

lcReturn = tcString
FOR lnI = LEN(tcString) TO 1 STEP -1
	IF ASC(SUBSTR(lcReturn,lnI,1)) < 32
		lcReturn = STUFF(lcReturn, lnI, 1, '')
	ENDIF 
ENDFOR 

RETURN lcReturn

ENDFUNC
*----------------------------------------------------------------------------------
FUNCTION WriteTable
LOCAL loWord, loDocument, lnI, lnJ, llWasOpen, lnError, lcOldOnError, llError, lnCurRow, lnCurCol, loRange, loTable
LOCAL lnRows, lnCols

lcOldOnError = ON("Error")
lnError = 0
ON ERROR lnError = ERROR()
*ON ERROR lnError = LINENO()
llError = .F.

IF ALEN(this.aTableData) = 1 AND EMPTY(this.aTableData[1])
	* There's no data to write out!
	ASSERT .F. message "No data in array to write!"
	this.cErrorMessage = "No data in array to write!"
	llError = .T.
ENDIF 

IF NOT llError 
	IF ISNULL(this.oDocument)
		IF EMPTY(this.cWordFile)
			this.cWordFile = GETFILE("doc")
			IF EMPTY(this.cWordFile)
				this.cErrorMessage = "No Word document specified"
				llError = .T.
			ENDIF
		ENDIF
		
		IF NOT llError AND NOT FILE(this.cWordFile)
			ASSERT .f. message this.cWordFile+" not found!"
			this.cErrorMessage = this.cWordFile+' not found!'
			llError = .T.
		ENDIF

		IF NOT llError
			loWord = CREATEOBJECT("Word.Application")
			IF VARTYPE(loWord) <> "O"
				ASSERT .F. MESSAGE "Couldn't instantiate Word"
				this.cErrorMessage = "Could not instantiate Word"
				llError = .T.
			ENDIF
			IF NOT llError
				loDocument = loWord.Documents.Open(this.cWordFile)
				IF VARTYPE(loDocument) <> "O"
					ASSERT .F. MESSAGE "Couldn't open document"
					this.cErrorMessage = "Could not open document"
					llError = .T.
				ENDIF 
			ENDIF 
			llWasOpen = .F.
		ENDIF
	ELSE
		loDocument = this.oDocument
		llWasOpen = .T.
	ENDIF
ENDIF

lnRows = ALEN(this.aTableData,1)+this.nStartRow-1
lnCols = ALEN(this.aTableData,2)+this.nStartCol-1

DO WHILE .T.
	IF llError OR lnError <> 0 
		EXIT
	ENDIF 
	IF loDocument.Tables.Count >= this.nTable
		EXIT 
	ENDIF
	
	* Create as many tables as necessary until we have the table number specified.
	IF NOT this.lCreateIfMissing
		ASSERT .F. message "Cannot find table number "+TRANSFORM(this.nTable)+" in document!"
		this.cErrorMessage = "Cannot find table number "+TRANSFORM(this.nTable)+" in document!"
		llError = .T.
	ELSE
		* Create a table at the end of the document
		loRange = loDocument.Range()
		* Move to the end of the document
		loRange.MoveEnd( 6 )
		loRange.Collapse( 0 )
		* Leave one empty line
		loRange.InsertAfter( CHR(13) )
		loRange.Collapse( 0 )
		loTable = loDocument.Tables.Add( loRange, lnRows, lnCols)
		IF VARTYPE(loTable) <> "O"
			ASSERT .F. message "Could not create table in document"
			this.cErrorMessage = "Could not create table in document"
			llError = .T.
		ENDIF 
	ENDIF
ENDDO

loTable = loDocument.Tables[this.nTable]
IF VARTYPE(loTable) <> "O"
	ASSERT .F. message "Cannot get object reference to table"
	this.cErrorMessage = "Cannot get object reference to table"
	llError = .T.
ENDIF

IF NOT llError AND lnError = 0
	loTable.AllowAutoFit = .T.
	IF loTable.Rows.Count < lnRows OR loTable.Columns.Count < lnCols
		IF this.lErrorIfSmaller
			ASSERT .F. message "Table is too small for the data array"
			this.cErrorMessage = "Table is too small for the data array"
			llError = .T.
		ELSE
			DO WHILE loTable.Rows.Count < lnRows
				loTable.Rows.Add
				IF lnError <> 0
					EXIT 
				ENDIF 
			ENDDO 
			DO WHILE loTable.Columns.Count < lnCols
				loTable.Columns.Add
				IF lnError <> 0
					EXIT 
				ENDIF 
			ENDDO 
		ENDIF 
	ENDIF
ENDIF 

IF NOT llError AND lnError = 0
	FOR lnI = this.nStartRow TO lnRows
		FOR lnj = this.nStartCol TO lnCols
			IF this.lShowStatus
				WAIT WINDOW NOWAIT "Writing row "+TRANSFORM(lnI)+" of "+TRANSFORM(lnRows)+ CHR(13) ;
						+"Column "+TRANSFORM(lnJ)+" of "+TRANSFORM(lnCols)
			ENDIF 
			
			lnCurRow = lnI - this.nStartRow + 1
			lnCurCol = lnJ - this.nStartCol + 1
			loTable.Rows[lnI].Cells[lnJ].Range.Text = this.aTableData[lnCurRow,lnCurCol]
			IF lnError <> 0
				EXIT 
			ENDIF 
		ENDFOR
	ENDFOR 			

	IF this.lShowStatus
		WAIT CLEAR
	ENDIF 
	loWord.Documents.Save
ENDIF

IF NOT llWasOpen
	IF VARTYPE(loDocument) = "O"
		loDocument.Close()
	ENDIF
	IF VARTYPE(loWord) = "O"
		loWord.Quit()
	ENDIF 
ENDIF 

ON ERROR &lcOldOnError
this.nErrorNum = lnError
IF lnError <> 0
	this.cErrorMessage = MESSAGE()
ENDIF 

ENDFUNC
*----------------------------------------------------------------------------------

ENDDEFINE 
***********************************************************************************
